module code.common.Stream where

open import Relation.Binary.PropositionalEquality
open import Data.Bool renaming (Bool to Bit; false to 0#; true to 1#)
open import Data.List renaming (map to mapL ; zip to zipL)
open import Data.List.NonEmpty hiding (map ; zip ; head)
open import Data.Product hiding (map ; zip)
open import Data.Nat using (ℕ; zero; suc)

open import code.common.Idioms

--
-- Streams
--

record Stream {ℓ} (A : Set ℓ) : Set ℓ where
  coinductive
  field hd : A
        tl : Stream A
open Stream public

record _≈_ {A} (xs ys : Stream A) : Set where
  coinductive
  field
    hd≈ : (hd xs ≡ hd ys)
    tl≈ : (tl xs ≈ tl ys)
open _≈_ public

-- alternative form.
record _≊_ {A} (xs ys : Stream A) : Set where
  coinductive
  field
    hdtl≊ : (hd xs ≡ hd ys) × (tl xs ≊ tl ys)
open _≊_ public

infix 5 _◁_

_◁_ : ∀ {a : Set} → a → Stream a → Stream a
hd (a ◁ _)  = a
tl (_ ◁ as) = as

unfoldStr : ∀ {s o : Set} → (s → o) → (s → s) → s → Stream o
hd (unfoldStr f g s) = f s
tl (unfoldStr f g s) = unfoldStr f g (g s)

repeatL⁺ : ∀ {a} → List⁺ a → Stream a
repeatL⁺ nzl = unfoldStr observe delta (y , ys , nzl)
  where
     y  = π₁ (Data.List.NonEmpty.uncons nzl)
     ys = π₂ (Data.List.NonEmpty.uncons nzl)
     observe : ∀ {a : Set} → (a × List a × List⁺ a) → a
     observe (a , _ , _) = a
     delta : ∀ {a} → (a × List a × List⁺ a) → (a × List a × List⁺ a)
     delta (a , (x ∷ xs) , nzl) = (x , xs , nzl)
     delta (a , [] , nzl)       = (z , zs , nzl)
        where
           z  = π₁ (Data.List.NonEmpty.uncons nzl)
           zs = π₂ (Data.List.NonEmpty.uncons nzl)

takeStr : ∀ {a : Set} → ℕ → Stream a → List a
takeStr zero str    = []
takeStr (suc n) str = (hd str) ∷ takeStr n (tl str)

refl≈ : ∀ {A} → (as : Stream A) → as ≈ as
hd≈ (refl≈ as) = refl 
tl≈ (refl≈ as) = refl≈ (tl as)

sym≈ : ∀ {A} → {as bs : Stream A} → as ≈ bs → bs ≈ as
hd≈ (sym≈ as≈bs) rewrite (hd≈ as≈bs) = refl
tl≈ (sym≈ as≈bs)                     = sym≈ (tl≈ as≈bs)

trans≈ : ∀ {A} → {as bs cs : Stream A} → as ≈ bs → bs ≈ cs → as ≈ cs
hd≈ (trans≈ as≈bs bs≈cs) rewrite (hd≈ as≈bs) = hd≈ bs≈cs
tl≈ (trans≈ as≈bs bs≈cs) = trans≈ (tl≈ as≈bs) (tl≈ bs≈cs)

{-
cong≈ : ∀ {a b} (f : Stream a → Stream b) { x y : Stream a } →
        x ≈ y → f x ≈ f y
hd≈ (cong≈ f x≈y) = {!!}        
tl≈ (cong≈ f x≈y) = {!!}        
-}

simp : ∀ {a : Set} → (x : a) → (t : Stream a) → tl (x ◁ t) ≡ t
simp = λ x t → refl

unfoldrule : ∀ {a c : Set} →
             (f : c → a)   →
             (g : c → c)   →
             (s : c)       →
             (unfoldStr f g s) ≈ ((f s) ◁ unfoldStr f g (g s))
hd≈ (unfoldrule f g s)                                          = refl
tl≈ (unfoldrule f g s) rewrite simp (f s) (unfoldStr f g (g s)) = refl≈ (unfoldStr f g (g s))


iterate : ∀ {W : Set} → (W → W) → W → Stream W
hd (iterate f x) = x
tl (iterate f x) = iterate f (f x)

map : ∀ {a b : Set} → (a → b) → Stream a → Stream b
map f = unfoldStr (f ∘ hd) tl

zip : ∀ {a b : Set} → Stream a → Stream b → Stream (a × b)
zip s1 s2 = let
               f = λ (s1 , s2) → (hd s1 , hd s2)
               g = λ (s1 , s2) → (tl s1 , tl s2)
             in
               unfoldStr f g (s1 , s2)

infix 4 _×ᶠ_

_×ᶠ_ : ∀ {a b c d : Set} →
      (f : a → b) →
      (g : c → d) →
      (a × c)     →
      (b × d)
f ×ᶠ g = λ (a , c) → (f a , g c)

gt0 : ℕ → Set
gt0 n = n Data.Nat.> 0

data Even : ℕ → Set where
  even-zero  : Even zero
  even-plus2 : {n : ℕ} → Even n → Even (suc (suc n))

_:>_ : ∀ {a : Set} → a → Stream a → Stream a
hd (a :> as) = a
tl (a :> as) = as

unzip₁ : ∀ {a b : Set} → Stream (a × b) → Stream a
hd (unzip₁ sab) = π₁ (hd sab)
tl (unzip₁ sab) = unzip₁ (tl sab)

unzip₂ : ∀ {a b : Set} → Stream (a × b) → Stream b
hd (unzip₂ sab) = π₂ (hd sab)
tl (unzip₂ sab) = unzip₂ (tl sab)

cycle : ∀ {a : Set} → a → List a → Stream a
hd (cycle a [])       = a
hd (cycle a (x ∷ as)) = a
tl (cycle a [])        = cycle a []
tl (cycle a (x ∷ as))  = cycle x (as ++ (a ∷ []))

-- length = 9
seed0 : List (Bit × Bit)
seed0 = (1# , 0#) ∷ (0# , 0#) ∷ (1# , 0#) ∷ (0# , 0#) ∷
        (1# , 0#) ∷ (0# , 0#) ∷ (1# , 0#) ∷ (0# , 0#) ∷ (1# , 0#) ∷ []

