module code.firrtl.TransCircuit where

-- This is the translation semantics into Device Calculus for Idealized FIRRTL.

open import Relation.Binary.PropositionalEquality hiding ([_])

open import Data.Bool renaming (Bool to Bit; false to 0#; true to 1#)
open import Data.Fin using (Fin; zero; suc)
open import Data.List hiding (lookup ; [_])
open import Data.Maybe hiding (map)
open import Data.Nat using (ℕ; _∸_; _<_ ; _≤_ ; _+_) renaming (zero to Z; suc to S)
open import Data.Product hiding (map ; swap ; curry ; uncurry)
open import Data.Vec hiding (foldl ; foldr ; _++_ ; map ; _[_]=_ ; length ; count) renaming (lookup to vecref)

open import code.common.Idioms hiding (π³⁴ ; π⁴⁴)
open import code.common.Contexts

open import code.firrtl.Types
open import code.firrtl.Syntax
open import code.firrtl.TransType

open import code.devicecalculus.Types
open import code.devicecalculus.BitPatterns
open import code.devicecalculus.Syntax

proj1 : ∀ {Γ a b} → Γ ⊩ᵈ ((a ⊗ b) ⇒ a)
proj1 = lam (π1 (var zero))

proj2 : ∀ {Γ a b} → Γ ⊩ᵈ ((a ⊗ b) ⇒ b)
proj2 = lam (π2 (var zero))

--
--
--

proj : ∀ {Γ a} → (t : Ty) → (i : a ∈ᵍ t) → Γ ⊩ᵈ (trTy t ⇒ (trTy (↑ a)))
proj ((↑ _) ⊗ _) Zᵍ  = lam (app proj1 (var zero))
proj (_ ⊗ ts) (Sᵍ i) = lam (app (proj ts i) (app proj2 (var zero)))

projTy : ∀ {Γ a} → (t : Ty) → (i : a ∈ᵗ t) → Γ ⊩ᵈ (trTy t ⇒ (trTy a))
projTy (_ ⊗ _) Zᵗ  = lam (app proj1 (var zero))
projTy (_ ⊗ ts) (Sᵗ i) = lam (app (projTy ts i) (app proj2 (var zero)))

--
-- Should form an n-tuple in device calculus
--

Gr⌈_⌉ : GrTy → Type
Gr⌈_⌉ = (trTy ∘ ↑_)

cxt⌈_⌉ : Cxt GrTy → Cxt Type
cxt⌈_⌉ = Data.List.map Gr⌈_⌉

conv∈ : ∀ {a} → (Γ : Cxt GrTy) → (a ∈ Γ) → Gr⌈ a ⌉ ∈ cxt⌈ Γ ⌉
conv∈ .(_ ∷ _) zero = zero
conv∈ (_ ∷ ys) (suc i₁) = suc (conv∈ ys i₁)

prj : ∀ {Γ a}            →
        (Δ : ModTy)        →
        (μ : Ty)           →
        (i : μ ∈ᵗ instsₒ Δ) →
        (j : a ∈ᵍ μ)  →
        Γ ⊩ᵈ (trTy (instsₒ Δ) ⇒ Gr⌈ a ⌉ )
prj Δ μ i j = lam (app (proj μ j) (app (projTy (instsₒ Δ) i) (var zero)))

fun× : ∀ {Γ i2 r2 o1 t ts}      →
       (Γ ⊩ᵈ i2 ⇒ r2 ⇒ o1 ⇒ t)  →
       (Γ ⊩ᵈ i2 ⇒ r2 ⇒ o1 ⇒ ts) →
       (Γ ⊩ᵈ i2 ⇒ r2 ⇒ o1 ⇒ t ⊗ ts)
fun× f1 f2 = lam (lam (lam
                   (app (app (app (rename push3 f1)
                         (var (suc (suc (zero))))) (var (suc zero))) (var zero)
                      ⊠
                      app (app (app (rename push3 f2) (var (suc (suc (zero))))) (var (suc zero))) (var zero))))

transE : ∀ {a} →
         (Γ : Cxt GrTy) →
         (Δ : ModTy)    →
         (Γ  ⊢ₓ a // Δ) →
         trTy (instsₒ Δ) ∷ trTy (regs Δ) ∷ trTy (ins Δ) ∷ cxt⌈ Γ ⌉ ⊩ᵈ trTy (↑ a)
transE Γ Δ (reg x)            = app (proj (regs Δ) x) (var (suc zero))
transE Γ Δ (inp x)            = app (proj (ins Δ) x) (var (suc (suc zero)))
transE Γ Δ (node x)           = var (suc (suc (suc (conv∈ Γ x))))
transE Γ Δ (qexp μ i j)       = app (prj Δ μ i j) (var zero)
transE Γ Δ (uint x)           = ⟨ x ⟩
transE Γ Δ (sint x)           = ⟨ x ⟩
transE Γ Δ (xtail m e)        = app (tlΛ m) (transE Γ Δ e)
transE Γ Δ (mux t e1 e2)      = caseᵇᵛ (transE Γ Δ t) (① ∷ [])
                                       (lam (rename push1 (transE Γ Δ e1)))
                                       (transE Γ Δ e2)
transE Γ Δ (_⊢ₓ_//_.and e1 e2) = transE Γ Δ e1 && transE Γ Δ e2
transE Γ Δ (_⊢ₓ_//_.add e1 e2) = transE Γ Δ e1 ⧾ transE Γ Δ e2

trB : (Γ : Cxt GrTy) →
      (Δ : ModTy)    →
      (t : Ty)       →
      (i : Γ ⊢ᵇ t // Δ) →
      cxt⌈ Γ ⌉ ⊩ᵈ (trTy (ins Δ) ⇒ trTy (regs Δ) ⇒ trTy (instsₒ Δ) ⇒ trTy t)
trB Γ Δ ((↑ t) ⊗ ts) (⊗case x i) = fun× (lam (lam (lam (transE Γ Δ x)))) (trB Γ Δ ts i)
trB Γ Δ Unit i                    = lam (lam (lam 𝟙))

trN : (Γ : Cxt GrTy) →
      (Δ : ModTy)    →
      (t : Ty)       →
      (i : Γ ⊢ⁿ t // Δ) →
      cxt⌈ Γ ⌉ ⊩ᵈ (trTy (ins Δ) ⇒ trTy (regs Δ) ⇒ trTy (instsₒ Δ) ⇒ trTy t)
trN Γ Δ t (node x b) = trB Γ Δ t b
   -- N.b., there's a dependency of b on x that is implicit in node rule.
   -- I.e., x isn't unused.

------------------------------------------------------
-- Module Rules
------------------------------------------------------

codomain : ModTy → Type
codomain Δ = trTy (outs Δ) ⊗ (trTy (regs Δ) ⊗ trTy (instsᵢ Δ))

---
--- Environments
---
--

-- A module, Μ ⊢ᵐ Δ, is translated to a closed term, [] ⊩ᵈ modelΔ Δ
-- This function (predicate, really) will be used to define device environments,
-- playing the analogous role of Value.
-- May need adjustment; e.g., do we expose all state types?
--

-- |
-- | Called inputs, registers, and outputs in the text.
-- |
inputs : ModTy → Type
inputs Δ = trTy (ins Δ)

registers : ModTy → Type
registers Δ = trTy (tystate Δ)

outputs : ModTy → Type
outputs Δ = trTy (outs Δ)

trΔ :  ModTy → Type
trΔ Δ = (i ⇒ s ⇒ (o ⊗ s)) ⊗ (i ⊗ s ⊗ o)
  where
     i = inputs Δ
     s = registers Δ
     o = outputs Δ
     
modelΔ :  ModTy → Set
modelΔ Δ = [] ⊩ᵈ trΔ Δ

ModEnv : Cxt ModTy → Set
ModEnv Μ = All modelΔ Μ

type𝟙 : Type
type𝟙 = (unit ⇒ unit ⇒ (unit ⊗ unit)) ⊗ (unit ⊗ unit ⊗ unit)

fun𝟙 : [] ⊩ᵈ type𝟙
fun𝟙 = (lam (lam (𝟙 ⊠ 𝟙))) ⊠ (𝟙 ⊠ (𝟙 ⊠ 𝟙))

ds2Type : Cxt ModTy → Type
ds2Type []       = type𝟙
ds2Type (d ∷ []) = trΔ d
ds2Type (d ∷ ds) = let
                     is : Type
                     is = trTy (Πty (ins d) (map ins ds))
                     os : Type
                     os = trTy (Πty (outs d) (map outs ds))
                     ms = trTy (Πty (tystate d) (tystates ds))
                   in
                     (is ⇒ (ms ⇒ (os ⊗ ms))) ⊗ (is ⊗ ms ⊗ os)

goofy : (x : ModTy) →
        (ds : Cxt ModTy) →
        [] ⊩ᵈ ds2Type (x ∷ ds) →
        [] ⊩ᵈ ((trTy (Πty (ins x) (map ins ds)))       ⇒
               ((trTy (Πty (tystate x) (tystates ds))) ⇒
               ((trTy (Πty (outs x) (map outs ds))) ⊗ (trTy (Πty (tystate x) (tystates ds))))))
goofy x [] txds        = π1 txds
goofy x (x₁ ∷ ds) txds = π1 txds

subgoofy : (i s o : Ty) → 
           (m : ModTy) →
           (μ : Cxt ModTy) →
           [] ⊩ᵈ ds2Type (m ∷ μ) →
           [] ⊩ᵈ trTy (instsᵢ (Mod i , s , o ⋊ m ∷ μ)) ⇒
                 Π (trTy (tystate m)) (map trTy (tystates μ)) ⇒
                 ((trTy (Πty (outs m) (map outs μ))) ⊗ Π (trTy (tystate m)) (map trTy (tystates μ)))
subgoofy i s o m [] txds      = π1 txds
subgoofy i s o m (x ∷ μ) txds rewrite lemma3 (ins m) (ins x ∷ map ins μ)
                                      | lemma3 (tystate m) (tystate x ∷ tystates μ) = π1 txds

getinit : (x : ModTy) →
          (ds : Cxt ModTy) →
          [] ⊩ᵈ ds2Type (x ∷ ds) →
          [] ⊩ᵈ trTy (Πty (ins x) (map ins ds))
                    ⊗
                trTy (Πty (tystate x) (tystates ds))
                    ⊗
                trTy (Πty (outs x) (map outs ds))
getinit x [] txds        = π2 txds
getinit x (x₁ ∷ ds) txds = π2 txds

help : ∀ {Μ} →
       (d x : ModTy) →
       (ds : Cxt ModTy) →
       ModEnv Μ      →
       (j : d ∈ Μ) →
       [] ⊩ᵈ ds2Type (x ∷ ds) →
          trTy (Πty (tystate d) (tystate x ∷ tystates ds)) ∷
          trTy (Πty (ins d) (ins x ∷ map ins ds)) ∷ []
             ⊩ᵈ trTy (Πty (outs d) (outs x ∷ map outs ds))
help d x [] ρ j tds =
                   π1 (app (app (rename push2 (π1 (lookup∈ ρ j)))
                                (π1 (var (suc zero))))
                                (π1 (var zero)))
                           ⊠
                   π1 (app (app (rename push2 (π1 tds))
                                (π2 (var (suc zero))))
                                (π2 (var zero)))

help d x ds ρ j tds
   rewrite lemma3 (outs d) (outs x ∷ map outs ds)
         | lemma1 (trTy (outs d)) (trTy (outs x)) (map trTy (map outs ds))
         | lemma3 (ins d) (map ins (x ∷ ds))
         | lemma1 (trTy (ins d)) (trTy (ins x)) (map trTy (map ins ds))
         | lemma4 (tystate d) (tystate x) (tystates ds)
         | lemma2 (ins x) (map ins ds)
         | lemma2 (tystate x) (tystates ds)
         | lemma2 (outs x) (map outs ds)
              = let
                  is = Πty (ins d) (map ins (x ∷ ds))
                  os = Πty (outs d) (map outs (x ∷ ds))
                  ms = Πty (tystate d) (tystates (x ∷ ds))
                  uf1 = lookup∈ ρ j
                  wtf = goofy x ds tds
                in
                   π1 (app (app (rename push2 (π1 uf1))
                                (π1 (var (suc zero))))
                                (π1 (var zero)))
                           ⊠
                   π1 (app (app (rename push2 wtf)
                                (π2 (var (suc zero))))
                                (π2 (var zero)))

helpme : ∀ {Μ} →
         (d x : ModTy) →
         (ds : Cxt ModTy) →
         ModEnv Μ      →
         (j : d ∈ Μ) →
         [] ⊩ᵈ ds2Type (x ∷ ds) →
         trTy (Πty (tystate d) (tystate x ∷ tystates ds)) ∷
         trTy (Πty (ins d) (ins x ∷ map ins ds)) ∷ []
           ⊩ᵈ trTy (Πty (tystate d) (tystate x ∷ tystates ds))
helpme d x [] ρ j tds        = (π2 (app (app (rename push2 (π1 (lookup∈ ρ j)))
                                             (π1 (var (suc zero))))
                                             (π1 (var zero))))
                                ⊠
                               π2 (app (app (rename push2 (π1 tds))
                                (π2 (var (suc zero))))
                                (π2 (var zero)))
helpme d x (x₁ ∷ ds) ρ j tds =
                   π2 (app (app (rename push2 (π1 (lookup∈ ρ j)))
                                (π1 (var (suc zero))))
                                (π1 (var zero)))
                           ⊠
                   π2 (app (app (rename push2 (π1 tds))
                                (π2 (var (suc zero))))
                                (π2 (var zero)))

unfold× : ∀ {Μ} →
          (d : ModTy) →
          (ds : Cxt ModTy) →
          ModEnv Μ      →
          (j : d ∈ Μ) →
          [] ⊩ᵈ ds2Type ds →
          [] ⊩ᵈ ds2Type (d ∷ ds)
unfold× d [] ρ j tds       = lookup∈ ρ j
unfold× d (x ∷ ds) ρ j tds =
  let
    is = Πty (ins d) (map ins (x ∷ ds))
    os = Πty (outs d) (map outs (x ∷ ds))
    ms = Πty (tystate d) (tystates (x ∷ ds)) 
  in
    lam (lam (help d x ds ρ j tds ⊠ helpme d x ds ρ j tds))
        ⊠
    ((value0 is) ⊠ (value0 ms ⊠ value0 os))

-- This function translates a non-empty μ into a single modelΔ.
transμ : (ds : List ModTy) →
         (Μ : Cxt ModTy) →
         ModEnv Μ        →
         Μ ⊢ᴹ ds   →
         [] ⊩ᵈ ds2Type ds
transμ [] Μ ρ m              = fun𝟙
transμ (d ∷ []) Μ ρ (x ◁ m)  = lookup∈ ρ x
transμ (d ∷ ds) Μ ρ (x ◁ ms) = unfold× d ds ρ x (transμ ds Μ ρ ms)

------
------
------

cnv : (i s o : Ty) →
      (m : ModTy) →
      (μ : Cxt ModTy) →
      [] ⊩ᵈ trTy (Πty (outs m) (map outs μ)) →
      [] ⊩ᵈ trTy (instsₒ (Mod i , s , o ⋊ m ∷ μ))
cnv i s o m [] x       = x
cnv i s o m (d ∷ μ) x rewrite lemma3 (outs m) (outs d ∷ map outs μ) = x

cnv' : (s : Ty) →
       (m : ModTy) →
       (μ : Cxt ModTy) →
       [] ⊩ᵈ trTy (prodTy (s ∷ tystate m ∷ tystates μ)) → 
       [] ⊩ᵈ trTy s ⊗ Π (trTy (tystate m)) (map trTy (tystates μ))
cnv' s m μ x rewrite lemma4 s (tystate m) (tystates μ) = x

trM : ∀ {Μ}      →
         (Δ : ModTy) →
         ModEnv Μ   →
         Μ ⊢ᵐ Δ     →
         [] ⊩ᵈ trΔ Δ     
trM Δ@(Mod i , s , o ⋊ []) δ (mod x o2 r2 i1) =
    let
      to2 = trN [] Δ o o2
      tr2 = trN [] Δ (regs Δ) r2
    in
      lam (lam (
             (app (app (app (rename push2 to2) (var (suc zero))) (var zero)) 𝟙)
                ⊠
             (app (app (app (rename push2 tr2) (var (suc zero))) (var zero)) 𝟙)))
        ⊠
      (value0 i ⊠ value0 s ⊠ value0 o)
trM {Μ} Δ@(Mod i , s , o ⋊ m ∷ μ) δ (mod ms o! s! mi!) 
    rewrite lemma4 s (tystate m) (tystates μ) =

    let
    
      to2 : [] ⊩ᵈ trTy i ⇒ trTy s ⇒ trTy (instsₒ (Mod i , s , o ⋊ m ∷ μ)) ⇒ trTy o
      to2 = trN [] Δ o o!

      tr2 : [] ⊩ᵈ trTy i ⇒ trTy s ⇒ trTy (instsₒ (Mod i , s , o ⋊ m ∷ μ)) ⇒ trTy s
      tr2 = trN [] Δ (regs Δ) s!

      tmi : [] ⊩ᵈ trTy i ⇒ trTy s ⇒ trTy (instsₒ (Mod i , s , o ⋊ m ∷ μ)) ⇒
                                                     trTy (instsᵢ (Mod i , s , o ⋊ m ∷ μ))
      tmi = trN [] Δ (instsᵢ Δ) mi!

      trms : [] ⊩ᵈ ds2Type (m ∷ μ)
      trms = transμ (m ∷ μ) Μ δ ms

      inits = getinit m μ trms
      mi : [] ⊩ᵈ trTy (Πty (ins m) (map ins μ))
      mi    = π1 inits
      mo : [] ⊩ᵈ trTy (Πty (outs m) (map outs μ))
      mo    = π2 (π2 inits)

      initstate = cnv' s m μ (value0 (prodTy (s ∷ tystate m ∷ tystates μ)))

    in

      lam (lam (app (app (app (rename push2 to2) (var (suc zero))) (π1 (var zero)))
                             (rename push2 (cnv i s o m μ mo))
                ⊠
                app (app (app (rename push2 tr2)
                              (var (suc zero)))
                              (π1 (var zero)))
                              (rename push2 (cnv i s o m μ mo))
                ⊠
                π2 (app (app (rename push2 (subgoofy i s o m μ trms))
                             (app (app (app (rename push2 tmi)
                                            (var (suc zero)))
                                            (π1 (var zero)))
                                            (rename push2 (cnv i s o m μ mo))))
                             (π2 (var zero)))))

        ⊠
      (value0 i ⊠ initstate ⊠ value0 o)

------------------------------------------------------
-- Circuit Rules
------------------------------------------------------

π¹⁴ : ∀ {Γ a b c d} → Γ ⊩ᵈ (a ⊗ b ⊗ c ⊗ d) → Γ ⊩ᵈ a
π¹⁴ x = π1 x

π²⁴ : ∀ {Γ a b c d} → Γ ⊩ᵈ (a ⊗ b ⊗ c ⊗ d) → Γ ⊩ᵈ b
π²⁴ x = π1 (π2 x)

π³⁴ : ∀ {Γ a b c d} → Γ ⊩ᵈ (a ⊗ b ⊗ c ⊗ d) → Γ ⊩ᵈ c
π³⁴ x = π1 (π2 (π2 x))

π⁴⁴ : ∀ {Γ a b c d} → Γ ⊩ᵈ (a ⊗ b ⊗ c ⊗ d) → Γ ⊩ᵈ d
π⁴⁴ x = π2 (π2 (π2 x))

trC : ∀ {Μ}      
    → (Δ : ModTy) 
    → ModEnv Μ   
    → Μ ⊢ᶜ Δ     
    → [] ⊩ᵈ Dev (inputs Δ) (registers Δ) (outputs Δ)
trC Δ ρ (mod x)     = unfoldDev (π¹⁴ ans) (π²⁴ ans) (π³⁴ ans) (π⁴⁴ ans)
  where
    ans = trM Δ ρ x
trC Δ ρ (letmod Δₓ x m) = trC Δ (trM Δₓ ρ x ▷ ρ) m
